package test.correctness;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Properties;

/**
 * 
 * The TPCBCorrectionCheck is a class responsible for checking the 
 * consistency in both the primary and the hot standby database. Both 
 * databases need to be started in the original server/client mode. 
 * The contents of the TPCB databases are then compared to eachother.
 * 
 * @author Egil Srensen
 *
 */

public class TPCBCorrectionCheck {
	/**
	 * Connection to the primary database
	 */
	private Connection primConn;
	/**
	 * Connection to the hot standby database
	 */
	private Connection hsbConn;
	/**
	 * Properties including username and password
	 */
	private Properties props;
	/**
	 * Statement used to execute sql queries to the primary database
	 */
	private Statement primS;
	/**
	 * Statement used to execute sql queries to the primary database
	 */
	private Statement hsbS;
	
	/**
	 * Main method to start the test
	 * 
	 * @param args	No arguments are used
	 */
	public static void main(String[] args){
		TPCBCorrectionCheck check = new TPCBCorrectionCheck();
		check.checkConsistency();
	}
	
	/**
	 * Initialize the test and set up connections
	 *
	 */
	public TPCBCorrectionCheck(){
        try {
			Class.forName("org.apache.derby.jdbc.ClientDriver").newInstance();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
        hsbConn = null;
        primConn = null;
        props = new Properties();
        props.put("user", "user1");
        props.put("password", "user1");
        try{
        	String primUrl = "lachdanan.egilnett.com";
        	String hsbUrl = "kratas.egilnett.com";
	        primConn = DriverManager.getConnection(
	        		"jdbc:derby://" + primUrl + ":1527/TPCB;create=true",
	        		props);

	        hsbConn = DriverManager.getConnection(
	        		"jdbc:derby://" + hsbUrl + ":1527/TPCB;create=true",
	        		props);
        } catch (SQLException e){
        	e.printStackTrace();
        }
        
	}
	
	
	
	/**
	 * Check the consistency of the two databases (primary and hot standby). The size 
	 * and contents of the tables BRANCHES, TELLERS, ACCOUNTS and HISTORY.
	 * 
	 * If any irregularities are found, the error is shown and the check is stopped.
	 *
	 */
	public void checkConsistency(){
		try {
			// Temporary variables
			String query;
			ResultSet primRs;
			ResultSet hsbRs;
			int primSize;
			int hsbSize;
			int diff;
			boolean error;
			int counter;

			// Create the statements
			primS = primConn.createStatement();
			hsbS = hsbConn.createStatement();			
			
			
			// Check the BRANCHES-table
			System.out.println("[BRANCHES]");
			error = false;


//			primS.executeUpdate("insert into branches (bid, bbalance, filler) values (2,0,'egil var her')");
			primS.executeUpdate("delete from branches where bid = 2");
			
			primS.executeUpdate("update tellers set tbalance = 2");
			
			// Getting size
			query = "select count(bid) as numBranches from branches";
			primRs = primS.executeQuery(query);
			hsbRs = hsbS.executeQuery(query);
			
			primRs.next();
			primSize = primRs.getInt("numBranches");
			hsbRs.next();
			hsbSize = hsbRs.getInt("numBranches");
			
			primRs.close();
			hsbRs.close();
			
			System.out.println("Checking size:");
			if (primSize == hsbSize)
				System.out.print("[OK] ");
			else{
				System.out.print("[ERROR] ");
				error = true;
			}
			System.out.print(primSize+":"+hsbSize+"\n");
			
			// Quit on error
			if (error)
				System.exit(-1);
			
			
			// Getting contents
			query = "select bid, bbalance, filler from branches order by bid asc";
			primRs = primS.executeQuery(query);
			hsbRs = hsbS.executeQuery(query);
			System.out.println("Checking contents ...");
			
			diff = 0;
			counter = 0;
			while (primRs.next() && hsbRs.next()){
				if ((primRs.getInt("bid") != hsbRs.getInt("bid")) || (primRs.getInt("bbalance") != hsbRs.getInt("bbalance"))){
					System.out.print("\n");
					diff++;
					System.out.println("[ERROR] bid: " + primRs.getInt("bid")+":"+hsbRs.getInt("bid")+", bbalance: " + primRs.getInt("bbalance") + ":" + hsbRs.getInt("bbalance"));
				}
				if (counter % 20000 == 0 && counter != 0)
					System.out.print("\n");
				if (counter % 500 == 0 && counter != 0)
					System.out.print(".");
				counter++;
			}
			System.out.print("\n");
			primRs.close();
			hsbRs.close();
			
			
			if (diff == 0){
				System.out.println("[OK]");
			}
			else{
				System.out.println("[ERROR] " + diff + " row(s) do(es) not match in the two databases. See above");
				System.exit(-1);
			}
			
			// Finished checking the BRANCHES table
			System.out.println("[/BRANCHES]");
			System.out.println("");
			error = false;
			
			// Check the TELLERS table
			System.out.println("[TELLERS]");
			

			// Getting size
			query = "select count(tid) as numTellers from tellers";
			primRs = primS.executeQuery(query);
			hsbRs = hsbS.executeQuery(query);
			
			primRs.next();
			primSize = primRs.getInt("numTellers");
			hsbRs.next();
			hsbSize = hsbRs.getInt("numTellers");
			primRs.close();
			hsbRs.close();

			System.out.println("Checking size:");
			if (primSize == hsbSize)
				System.out.print("[OK] ");
			else{
				System.out.print("[ERROR] ");
				error = true;
			}
			
			System.out.print(primSize+":"+hsbSize+"\n");
			
			// Quit on error
			if (error)
				System.exit(-1);
			
			// Getting contents
			query = "select tid, bid, tbalance, filler from tellers order by tid asc";
			primRs = primS.executeQuery(query);
			hsbRs = hsbS.executeQuery(query);
			System.out.println("Checking contents ...");
			
			diff = 0;
			counter = 0;
			while (primRs.next() && hsbRs.next()){
				if ((primRs.getInt("tid") != hsbRs.getInt("tid")) || (primRs.getInt("bid") != hsbRs.getInt("bid")) || (primRs.getInt("tbalance") != hsbRs.getInt("tbalance"))){
					diff++;
					System.out.print("\n");
					System.out.println("[ERROR] tid: " + primRs.getInt("tid")+":"+hsbRs.getInt("tid")+", bid: " + primRs.getInt("bid")+":"+hsbRs.getInt("bid")+", tbalance: " + primRs.getInt("tbalance") + ":" + hsbRs.getInt("tbalance"));
				}
				if (counter % 20000 == 0 && counter != 0)
					System.out.print("\n");
				if (counter % 500 == 0 && counter != 0)
					System.out.print(".");
				counter++;
			}
			System.out.print("\n");
			primRs.close();
			hsbRs.close();
			
			if (diff == 0){
				System.out.println("[OK]");
			}
			else{
				System.out.println("[ERROR] " + diff + " row(s) do(es) not match in the two databases. See above");
				System.exit(-1);
			}
			
			// Finished checking the TELLERS table
			System.out.println("[/TELLERS]");			
			System.out.println("");
			error = false;
			
			// Checking the ACCOUNTS table
			System.out.println("[ACCOUNTS]");

			// Getting size
			query = "select count(aid) as numAccounts from accounts";
			primRs = primS.executeQuery(query);
			hsbRs = hsbS.executeQuery(query);
			
			primRs.next();
			primSize = primRs.getInt("numAccounts");
			hsbRs.next();
			hsbSize = hsbRs.getInt("numAccounts");
			primRs.close();
			hsbRs.close();
			

			System.out.println("Checking size:");
			if (primSize == hsbSize)
				System.out.print("[OK] ");
			else{
				System.out.print("[ERROR] ");
				error = true;
			}
			
			System.out.print(primSize+":"+hsbSize+"\n");
			
			// Quit on error
			if (error)
				System.exit(-1);
			
			// Getting contents
			query = "select aid, bid, abalance, filler from accounts order by aid asc";
			primRs = primS.executeQuery(query);
			hsbRs = hsbS.executeQuery(query);
			System.out.println("Checking contents ...");
			
			diff = 0;
			counter = 0;
			while (primRs.next() && hsbRs.next()){
				if ((primRs.getInt("aid") != hsbRs.getInt("aid")) || (primRs.getInt("bid") != hsbRs.getInt("bid")) || (primRs.getInt("abalance") != hsbRs.getInt("abalance"))){
					System.out.print("\n");
					diff++;
					System.out.println("[ERROR] aid: " + primRs.getInt("aid")+":"+hsbRs.getInt("aid")+", bid: " + primRs.getInt("bid")+":"+hsbRs.getInt("bid")+", abalance: " + primRs.getInt("abalance") + ":" + hsbRs.getInt("abalance"));
				}
				if (counter % 20000 == 0 && counter != 0)
					System.out.print("\n");
				if (counter % 500 == 0 && counter != 0)
					System.out.print(".");
				counter++;
			}
			System.out.print("\n");
			primRs.close();
			hsbRs.close();
			
			if (diff == 0){
				System.out.println("[OK]");
			}
			else{
				System.out.println("[ERROR] " + diff + " row(s) do(es) not match in the two databases. See above");
				System.exit(-1);
			}
			
			// Finished checking the ACCOUNTS table
			System.out.println("[/ACCOUNTS]");				
			System.out.println("");
			error = false;
			
			// Check the HISTORY table
			System.out.println("[HISTORY]");
			

			// Getting size
			query = "select count(tid) as numHistory from history";
			primRs = primS.executeQuery(query);
			hsbRs = hsbS.executeQuery(query);
			
			primRs.next();
			primSize = primRs.getInt("numHistory");
			hsbRs.next();
			hsbSize = hsbRs.getInt("numHistory");
			primRs.close();
			hsbRs.close();
			

			System.out.println("Checking size:");
			if (primSize == hsbSize)
				System.out.print("[OK] ");
			else{
				System.out.print("[ERROR] ");
				error = true;
			}
			
			System.out.print(primSize+":"+hsbSize+"\n");
			
			// Quit on error
			if (error)
				System.exit(-1);
			
			// Getting contents
			query = "select aid, bid, tid, delta, time, filler from history order by tid, bid, aid, time asc";
			primRs = primS.executeQuery(query);
			hsbRs = hsbS.executeQuery(query);
			System.out.println("Checking contents ...");
			
			diff = 0;
			counter = 0;
			while (primRs.next() && hsbRs.next()){
				if ((primRs.getInt("tid") != hsbRs.getInt("tid")) || (primRs.getInt("bid") != hsbRs.getInt("bid")) || (primRs.getInt("aid") != hsbRs.getInt("aid")) || (primRs.getInt("delta") != hsbRs.getInt("delta")) || ! primRs.getTime("time").equals(hsbRs.getTime("time"))){
					System.out.print("\n");
					diff++;
					System.out.println("[ERROR] aid: " + primRs.getInt("aid")+":"+hsbRs.getInt("aid")+", tid: " + primRs.getInt("tid")+":"+hsbRs.getInt("tid")+", bid: " + primRs.getInt("bid") + ":" + hsbRs.getInt("bid")+", delta: " + primRs.getInt("delta") + ":" + hsbRs.getInt("delta")+", time: " + primRs.getTime("time") + ":" + hsbRs.getTime("time"));
				}
				if (counter % 20000 == 0 && counter != 0)
					System.out.print("\n");
				if (counter % 500 == 0 && counter != 0)
					System.out.print(".");
				counter++;
			}
			System.out.print("\n");
			primRs.close();
			hsbRs.close();
			
			if (diff == 0){
				System.out.println("[OK]");
			}
			else{
				System.out.println("[ERROR] " + diff + " row(s) do(es) not match in the two databases. See above");
				System.exit(-1);
			}
			
			// Finished checking the HISTORY table
			System.out.println("[/HISTORY]");				
			
			// Close the resultsets
			primRs.close();
			hsbRs.close();
			
			// Close statements and connections
			primS.close();
			hsbS.close();
			primConn.close();
			hsbConn.close();

		} catch (SQLException e) {
			e.printStackTrace();
		}
	}

	
	
	
	
}
