package org.apache.derby.impl.drda;

import org.apache.derby.impl.drda.net.LogicalClient;
import org.apache.derby.impl.drda.net.LogicalServer;
import org.apache.derby.impl.services.hotstandby.HotStandbyController;
import org.apache.derby.impl.services.net.NetworkPayload;

/**
 * The HotStandbyPoller is the class responsible for sending I'm-alive-messages 
 * to the HotStandby to ensure that the system can fail-fast when the primary is no 
 * longer alive. When no I'm-alive-messages are received in ten seconds the primary 
 * is assumed dead and the HotStandby performs a takeover. When the takeover 
 * is completed, the new primary attempts to reconnect to the new HotStandby and 
 * send I'm-alive-messages for as long as it is alive.
 * 
 * @author Egil Srensen
 *
 */

public class HotStandbyPoller implements Runnable {

	// The instance of the poller
	private static HotStandbyPoller instance;
	// The server used to receive I'm-alive-messages
	private LogicalServer server;
	// The client used to send I'm-alive-messages
	private LogicalClient client;
	// Is the database in HotStandbyMode?
	private boolean hsb;
	// How long the polling intervals are
	private long pollingInterval = 100;
	// How long a timeout should take before the primary 
	// is assumed dead. Set to 10 seconds
	private long timeoutInterval = 10000;
	// The time left until the next timeout
	private long nextTimeOut;
	// The owner of this Poller. Used to signal when the HotStandby
	// needs to change its mode to Primary
	private NetworkServerControlImpl owner;
	
	
	/**
	 * Get the instance of the poller
	 * 
	 * @return		The poller instance
	 */
	public static HotStandbyPoller getInstance(){
		if (instance == null)
			instance = new HotStandbyPoller();
		return instance;
	}
	
	/**
	 * Private empty constructor
	 *
	 */
	
	private HotStandbyPoller(){
		
	}
	
	/**
	 * Initialize the poller, start the server if the database is in HotStandbyMode, if 
	 * not, start the client
	 * 
	 * @param owner		The owner of this poller
	 * @param hsb		Is the database in HotStandbyMode?
	 */
	protected void init(NetworkServerControlImpl owner, boolean hsb){
		String url = HotStandbyController.getInstance().getHotStandbyUrl();
		this.hsb = hsb;
		this.owner = owner;
		if (hsb){
			startServer();
		}
		else{
			startClient(url);
		}
		new Thread(this).start();
	}
	
	
	/**
	 * Initializes the network server used for I'm-alive-messages
	 *
	 */
	private void startServer(){
		LogicalServer.init();
		server = LogicalServer.getInstance();
		notifyAlive();
	}
	
	/**
	 * Initializes the network client used to ship I'm-alive-messages
	 * 
	 * @param url		The url to the HotStandby
	 */
	private void startClient(String url){
		client = LogicalClient.getInstance(url);
	}
	
	
	/**
	 * Notify the poller that the primary is still alive, reset the nextTimeout-timer
	 *
	 */
	public void notifyAlive(){
		nextTimeOut = System.currentTimeMillis() + timeoutInterval;
	}
	
	
	/**
	 * Runs the poller
	 */
	public void run() {
		
		// As long as the database is in HotStandbyMode, listen for I'm-alive-messages
		while (hsb) {
			if (System.currentTimeMillis() > nextTimeOut){
				System.out.println("Timeout detected !! Primary is no longer alive ?!?");
				hsb = !hsb;
				String url = server.getClientId();
				HotStandbyController.getInstance().setHotStandbyUrl(url);
				owner.setHotStandby(hsb);
				server.shutdown();
				startClient(url);
				
			}
			else{
				try {
					Thread.sleep(pollingInterval);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}

		// As long as the database is in PrimaryMode, ship I'm-alive-messages
		while (!hsb) {
			client.send(new NetworkPayload(NetworkPayload.TYPE_IMALIVE,
					System.currentTimeMillis() + ""));

			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}	
	}


}
