package org.apache.derby.impl.drda.net;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.ConnectException;
import java.net.Socket;
import java.net.UnknownHostException;

import org.apache.derby.impl.services.hotstandby.HotStandbyController;
import org.apache.derby.impl.services.net.NetworkPayload;
import org.apache.derby.impl.services.net.shipper.LogShipper;
import org.apache.derby.impl.store.logical.log.LogicalLogger;

/**
 * LogicalClient is a network client capable of delivering NetworkPayloads 
 * to a LogicalServer on another database. It supplies the database with an 
 * easy-to-use request-response network protocol.
 * 
 * The LogicalClient is also responsible for keeping the database notified 
 * if the HotStandby is alive or not.
 * 
 * @author Egil Srensen
 *
 */
public class LogicalClient {

	// The instance of the client
	private static LogicalClient instance;
	// The stream used to write the request-payload to
	private ObjectOutputStream oos;
	// The stream used to read the response-payload from
	private ObjectInputStream ois;
	// The socket containing the connection to the server
	private Socket networkConnection;
	// The url to the server
	private String logServer;
	// The port to the server
	private int logServerPort = 12346;
	
	/**
	 * Get an instance for this client. If an instance exists the 
	 * url is set to the chosen url and the client reconnects. If not 
	 * a new LogicalClient is initialized and returned 
	 * 
	 * @param url		The url to the LogicalServer
	 * @return
	 */
	public static LogicalClient getInstance(String url){
		if (instance == null)
			instance = new LogicalClient(url);
		else
			instance.setUrl(url);
		return instance;
	}
	
	
	/**
	 * Initialize a new LogicalClient
	 * 
	 * @param url		The url to the LogicalServer
	 */
	private LogicalClient(String url){
		setUrl(url);
	}
	
	/**
	 * Set the url of the LogicalServer and reconnect
	 * 
	 * @param url		The url to the LogicalServer
	 */
	private void setUrl(String url){
		logServer = url;
		reconnect();
	}
	
	
	
	/**
	 * Send a NetworkPayload and return the response from the LogicalServer
	 * 
	 * @param payload		The request-payload
	 * @return				The response-payload
	 */
	
	public synchronized NetworkPayload send(NetworkPayload payload){
		NetworkPayload answer = null;

		// Try to send the networkpayload by using the ObjectOutputStream
		try {
			oos.reset();
			oos.writeObject(payload);
			oos.flush();
			answer = (NetworkPayload) ois.readObject();
			
		} catch (IOException e) {
			// If there is a network problem, reconnect
			System.out.println("[warning] connection seems lost ... retrying ...");
			oos = null;
			ois = null;
			reconnect();
			answer = new NetworkPayload(NetworkPayload.TYPE_ERROR, "Connection lost");
		} catch (ClassNotFoundException e){
			// The returned class is not a NetworkPayload. This should never happen
			System.out.println("[warning] returned class not found ... " + e.getMessage());
		}
		
		return answer;
		
		
		
	}
	
	
	
	/**
	 * Reconnect to the LogicalServer and create the ObjectOutputStream needed 
	 * for the LogicalClient
	 *
	 */
	private synchronized void reconnect(){
		HotStandbyController.getInstance().setHotStandByAlive(false);
		System.out.println("[status] LogShipper is connecting to: " + logServer + ":" + logServerPort);
		try {
			// Create the connection
			networkConnection = new Socket(logServer, logServerPort);
			networkConnection.setKeepAlive(true);
			oos = new ObjectOutputStream(networkConnection.getOutputStream());
			ois = new ObjectInputStream(networkConnection.getInputStream());
			LogicalLogger.getInstance();
			// Provide catchup for the HotStandby if needed
			if (LogShipper.getInstance() == null)
				LogicalLogger.getInstance();
			LogShipper.getInstance().provideCatchUp();
		} catch (ConnectException e){
			// If the connection was refused, retry after a random number of seconds between 1 to 10
			long randomTime = (int) (Math.random() * 5000);
			System.out.println("[error] Could not connect: " + e.getMessage() + ". Retrying in " + (randomTime / 1000) + " seconds");
			
			try{
				Thread.sleep(randomTime);
			}
			catch (InterruptedException e1){
				System.out.println("[alert] Thread was intterrupted ... " + e1.getMessage());
			}
			// Try to reconnect again
			reconnect();
		} catch (UnknownHostException e1) {
			e1.printStackTrace();
		} catch (IOException e1) {
			e1.printStackTrace();
		}
	}

	
	
}
