package org.apache.derby.impl.drda.net;

import java.io.EOFException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.ServerSocket;
import java.net.Socket;

import org.apache.derby.impl.drda.HotStandbyPoller;
import org.apache.derby.impl.services.hotstandby.HotStandbyController;
import org.apache.derby.impl.services.net.NetworkPayload;
/**
 * The Logical Server provides, along with the LogicalClient, an easy-to-use 
 * request-response protocol. The Logical Server in the drda-package (Network 
 * Server) is responsible for receiving I'm-alive-messages to enable fail-fast 
 * semantics in the HotStandby-scheme
 * 
 * @author Egil Srensen
 *
 */

public class LogicalServer extends Thread{

	// The instance of the LogicalServer
	private static LogicalServer instance;
	// The stream to read requests from
	private ObjectInputStream ois;
	// The stream to write responses on
	private ObjectOutputStream oos;
	// The socket / connection to the client
	private ServerSocket socket;
	// The port the server should be listening on
	private int serverPort = 12346;
	// The url to the client
	private String clientId;
	// Should the server shut down?
	private boolean shutdown = false;
	
	/**
	 * Initialize and start the server
	 *
	 */
	public static void init(){
		if (instance == null){
			instance = new LogicalServer();
			new Thread(instance).start();
		}
	}
	
	/**
	 * Get an instance of this server
	 * 
	 * @return		The server instance
	 */
	public static LogicalServer getInstance(){
		return instance;
	}
	
	/**
	 * Private empty constructor
	 *
	 */
	private LogicalServer(){
		
	}
	
	/**
	 * Get the url of the client connected to the server
	 * 
	 * @return		The id or url belonging to the client currently connected
	 */
	public String getClientId(){
		return clientId;
	}
	
	/**
	 * Shutdown the server
	 *
	 */
	public void shutdown(){
		try {
			socket.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		shutdown = true;
	}
	
	
	
	/**
	 * Run the server
	 */
	public void run() {
		socket = null;
		try {
			// Initialize the server
			socket = new ServerSocket(serverPort);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		
		while (!shutdown){
			boolean running;
			Socket client = null;
			// Start the server and wait for connections
			System.out.println("[wait] Waiting for connections ...");
			try {
				client = socket.accept();
				client.setKeepAlive(true);
				clientId = client.getInetAddress().getHostAddress();
				System.out.println("[connection] Got a connection from " + clientId);
				ois = new ObjectInputStream(client.getInputStream());
				oos = new ObjectOutputStream(client.getOutputStream());
				running = true;
				HotStandbyController.getInstance().setHotStandbyUrl(clientId);

			} catch (IOException e) {
				e.printStackTrace();
				running = false;
			}
			
			// While the client is connected, get the request and form an appropriate response.
			// This LogicalServer is only responsible for sending ACKS in response to 
			// I'm-alive-messages
			while (running && !shutdown){
				try {

					// Get the incoming log
					NetworkPayload payload = (NetworkPayload) ois.readObject();
					
					NetworkPayload answer;
					
					switch (payload.getType()){
					case NetworkPayload.TYPE_IMALIVE:{
						answer = new NetworkPayload(NetworkPayload.TYPE_ACK, "ACK I'm Alive");
						HotStandbyPoller.getInstance().notifyAlive();
						break;	
					}
					default:{
						answer = new NetworkPayload(NetworkPayload.TYPE_ERROR,"Unkown request: " + payload.getType());
						break;
					}
					}
					
					oos.reset();
					oos.writeObject(answer);
					oos.flush();
					
					
					
					
				} catch (EOFException e){
					// The connection was broken, flush to disk and go back to waiting for connections
					System.out.println("[error] EOF detected ... connection shut down ?");
					running = false;
				} catch (IOException e) {
					e.printStackTrace();
				} catch (ClassNotFoundException e){
					e.printStackTrace();
				}
			}
		}
		
		
		
	}
	
	
	
	
}
