package org.apache.derby.impl.services.hotstandby;


import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

import org.apache.derby.iapi.store.logical.log.LogicalLoggable;
import org.apache.derby.impl.services.net.LogicalServer;
import org.apache.derby.impl.services.net.receiver.LogReceiver;
import org.apache.derby.impl.store.logical.log.LogicalLog;
import org.apache.derby.impl.store.logical.log.LogicalLogRecord;

/**
 * 
 * The HotStandbyController is the main class responsible for keeping the 
 * HotStandby connected and up-to-date. When the database is in Hot Standby Mode 
 * it is responsible for redoing incoming log records. During takeover it is 
 * also responsible for aborting running transactions and ensure that the database 
 * is in a consistent state before it goes Primary.
 * 
 * @author Egil Srensen
 *
 */
public class HotStandbyController {

	// The instance of the HotStandbyController
	private static HotStandbyController instance;
	// Is the database in HotStandbyMode?
	private boolean isInHotStandbyMode;
	// Is the HotStandby alive?
	private boolean hotStandByAlive;
	// A table with the current running transactions
	private Hashtable runningTransactions;
	// A connection to the database
	private Connection conn;
	// The url to the HotStandby
	private String hotStandbyUrl;
	// The selected replication mode, can be either ONE_SAFE or TWO_SAFE
	private int replicationMode;
	// Redo deamon used to redo incoming log elements
	private RedoDaemon redoDaemon;
	
	// Constants
	public final static int ONE_SAFE = 1;
	public final static int TWO_SAFE = 2;

	
	/**
	 * Initialize the HotStandbyController and the LogicalServer
	 * 
	 * @param isInHotStandbyMode	Is the database in HotStandbyMode?
	 */
	public static void init(boolean isInHotStandbyMode){
		if (instance == null){
			instance = new HotStandbyController(isInHotStandbyMode);
			LogicalServer.init();
		}
	}
	
	/**
	 * Get the running instance of the HotStandbyController
	 * 
	 * @return		The active HotStandbyController
	 */
	public static HotStandbyController getInstance(){
		if (instance == null)
			init(false);
		return instance;
	}
		
	
	/**
	 * Private constructor for HotStandbyController. Initializes 
	 * with HotStandbyController in the selected HotStandbyMode. 
	 * The HotStandby is assumed offline until discovered by the 
	 * I'm-alive-service implemented in the NetworkServer
	 * 
	 * @param isInHotStandbyMode
	 */
	private HotStandbyController(boolean isInHotStandbyMode){
		this.isInHotStandbyMode = isInHotStandbyMode;
		hotStandByAlive = false;
		replicationMode = ONE_SAFE;
//		if (isInHotStandbyMode){
			redoDaemon = new RedoDaemon();
			runningTransactions = new Hashtable();
//		}
	}
	
	
	/**
	 * Is the database in HotStandbyMode ?
	 * 
	 * @return
	 */
	public boolean isInHotStandbyMode(){
		return isInHotStandbyMode;
	}
	
	
	/**
	 * Set the databases HotStandbyMode. If the database is switched to 
	 * false, the HotStandbyController is forced to abort all running 
	 * transactions before the mode is set to false
	 * 
	 * @param toggle		The mode to be set
	 */
	public synchronized void setHotStandbyMode(boolean toggle){
		if (! toggle){
			redoDaemon.shutdown();
		}
		isInHotStandbyMode = toggle;
	}
	
	
	/**
	 * Abort all the running transactions registered in the 
	 * runningTransactions hashtable. The operations for these transactions 
	 * are rolled back using log scans and the undoOperations registered for these 
	 * transactions.
	 *
	 */
	private synchronized void abortAllRunningTransactions(){
		if (runningTransactions.size() > 0){
			connectDb();
//			System.out.println("[status] Commensing clean-up of still running transactions...");
			Enumeration keys = runningTransactions.keys();
			while (keys.hasMoreElements()){
				Integer key = (Integer) keys.nextElement();
				int transid = key.intValue();
				int lastLsn = ((Integer) runningTransactions.get(key)).intValue();
				abortTransaction(transid, lastLsn);
			}
			runningTransactions = new Hashtable();
//			System.out.println("[done]");
			disconnectDb();
		}
	}
	
	
	/**
	 * Connect to the internal database
	 *
	 */
	private void connectDb(){
		try {
			String dbName = "TPCB";
			Database db = new Database(dbName, "user1", "user1");
			conn = db.getConnection();
		} catch (SQLException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Disconnect from the internal database
	 *
	 */
	private void disconnectDb(){
		try {
			conn.close();
		} catch (SQLException e) {
			e.printStackTrace();
		}
	}
	
	
	
	/**
	 * Handle incoming logs from the LogReceiver. The log is scanned from the 
	 * start and operations are redone until all are done. The log is then flushed 
	 * to disk. While the operations are redone the runningTransaction-hashtable 
	 * is updated. When an operation is committed or aborted they are removed 
	 * from the runningTransaction-table.
	 * 
	 * @param log		The incoming log to be redone
	 */
	public synchronized void handleIncomingLogs(LogicalLog log){
		redoDaemon.redoLog(log);
	}
	
	
	/**
	 * Abort the selected transaction and roll back all operations 
	 * done by this transaction
	 * 
	 * @param transactionId		The id of the transaction to be aborted
	 * @param lastLsn			The last lsn affected by the transaction to be aborted
	 */
	private synchronized void abortTransaction(int transactionId, int lastLsn){
//		System.out.println("[abort] Aborting transaction " + transactionId);
		// Rewind to the first operation to be aborted
		try{
			LogicalLogRecord currentAbortRecord = LogReceiver.getInstance().getLog().rewind(lastLsn);
			if (currentAbortRecord != null){
				try {
					// Undo this operation.
					currentAbortRecord.getOperation().undoMe(conn);
				} catch (SQLException e) {
	//				e.printStackTrace();
				}
			
				// As long as the operation is not the first of the transaction rewind to 
				// the previous operation and undo it
				while (currentAbortRecord.getPrevTransLsn() != -1){
					currentAbortRecord = LogReceiver.getInstance().getLog().rewind(currentAbortRecord.getPrevTransLsn());
			
					try {
						currentAbortRecord.getOperation().undoMe(conn);
					} catch (SQLException e) {
	//					e.printStackTrace();
					}
				}
			}
		}
		catch (NullPointerException e){

		}
	}

	
	/**
	 * Get the url for the HoStandby
	 * 
	 * @return		Url for the HotStandby
	 */
	public String getHotStandbyUrl() {
		return hotStandbyUrl;
	}

	/**
	 * Set the HotStandby url
	 * 
	 * @param 		hotStandbyUrl
	 */
	public void setHotStandbyUrl(String hotStandbyUrl) {
		this.hotStandbyUrl = hotStandbyUrl;
	}

	
	/**
	 * Is the HotStandby alive?
	 * 
	 * @return		Is the HotStandby alive?
	 */
	public boolean isHotStandByAlive() {
		return hotStandByAlive;
	}
	

	/**
	 * Set the HotStandby alive or not
	 * 
	 * @param hotStandByAlive	Is the HotStandby alive?
	 */
	public void setHotStandByAlive(boolean hotStandByAlive) {
//		System.out.println("Setting hot standby alive ? " + hotStandByAlive);
		this.hotStandByAlive = hotStandByAlive;
	}

	
	/**
	 * Get the replication mode
	 * 
	 * @return 		The replication mode, ONE_SAFE or TWO_SAFE
	 */
	public int getReplicationMode() {
		return replicationMode;
	}

	/**
	 * Set the replication mode
	 * 
	 * @param replicationMode		The mode to be set, must be either ONE_SAFE or TWO_SAFE
	 */
	public void setReplicationMode(int replicationMode) {
		this.replicationMode = replicationMode;
	}
	
	
	
	
	/**
	 * TODO: javadoc denne om den virker :)
	 * Burde f opp throughput og forbedre responstid drastisk
	 * 
	 * Potensielt problem: loggen blir fylt opp fortere enn den kan omgjres, m finne p 
	 * noe fornuftig for  unng disse situasjonene. Bremse ned, la hotstandbycontroller 
	 * vente i redoLog eller noe
	 * 
	 * @author Egil Srensen
	 *
	 */
	private class RedoDaemon implements Runnable{
		
		private boolean shutdown;
		private ArrayList redoList;
		private final int MAX_LIST_SIZE = 40;
		
		public RedoDaemon(){
			shutdown = false;
			redoList = new ArrayList();
			new Thread(this, "RedoDaemon").start();
		}
		
		public synchronized void redoLog(LogicalLog redoLog){
//			System.out.println("Handling incoming log ... " + redoLog.size());
//			this.redoLog.addElements(redoLog);
			redoList.add(redoLog.flush());
			System.out.println("[+] " + redoList.size() + "/" + MAX_LIST_SIZE);
//			System.out.println("Redo-log is now: " + this.redoLog.size());
			if (redoList.size() >= MAX_LIST_SIZE){
				LogicalServer.getInstance().setWait(true);
			}
		}

		public void run() {
			while (! shutdown  && isInHotStandbyMode){
//				System.out.println("gogogo");
				int totalOperations = 0;
				int affectedRows = 0;
				// Connect to the database
				long totalRunTime = 0;
				int numRuns = 0;
				while (redoList.size() > 0){
					connectDb();
					long timestamp = System.currentTimeMillis();

					
					LogicalLog tempLog = (LogicalLog) redoList.remove(0);
					
					LogReceiver.getInstance().getLog().addElements(tempLog);
					tempLog.reset();
					// Go through the log
					while (tempLog.hasMoreElements() && tempLog.size() > 0){
						
						LogicalLogRecord record = tempLog.getNext();
						LogicalLoggable operation = record.getOperation();
						// Updating the runningTransactions-table
						runningTransactions.put(new Integer(record.getTransId()), new Integer(record.getLsn()));
						// The operation is neither commit or abort. Redo this operation.
						if (operation.group() != LogicalLoggable.COMMIT_OPERATION && operation.group() != LogicalLoggable.ABORT_OPERATION){
							try {
								totalOperations++;
								affectedRows += operation.redoMe(conn);
							} catch (SQLException e) {
								e.printStackTrace();
							}
						}
						// The operation is an abort, abort this
						else if (operation.group() == LogicalLoggable.ABORT_OPERATION){
							int lastLsn = record.getLsn();
							int transId = record.getTransId();
							abortTransaction(transId, lastLsn);
						}
						// Finally, if the operation is abort or commit the transaction is removed from runningTransactions.
						if (operation.group() == LogicalLoggable.ABORT_OPERATION || operation.group() == LogicalLoggable.COMMIT_OPERATION && runningTransactions.containsKey(new Integer(record.getTransId()))){
							runningTransactions.remove(new Integer(record.getTransId()));
						}
					}
					
					
					// Disconnect from the database
					disconnectDb();
	
//					System.out.println("[done] " + affectedRows + "/" + totalOperations + " rows were affected by batch update");
//					System.out.println("[status] Number of still running transactions: " + runningTransactions.size());
					
					affectedRows = 0;
					totalOperations = 0;
					
					LogReceiver.getInstance().getLog().flushLogToDisk();
//					if (redoList.size() < 8)
					totalRunTime += (System.currentTimeMillis() - timestamp);
					
					System.out.println("[-] " + Math.round((totalRunTime / ++numRuns)));
					LogicalServer.getInstance().setWait(false);
				}

				
				// Try to wait for 0,1 seconds before retrying
				try {
					Thread.sleep(100);
				} catch (InterruptedException e) {
					// Thread interrupted, should not happen
					e.printStackTrace();
				}
				LogReceiver.getInstance().getLog().flushLogToDisk();
				
			}
			
			abortAllRunningTransactions();
			
		}
		
		public void shutdown(){
			shutdown = true;
		}
	
	}
	
	
	
	
	
}
