package org.apache.derby.impl.services.net;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.ConnectException;
import java.net.Socket;
import java.net.UnknownHostException;

import org.apache.derby.impl.services.hotstandby.HotStandbyController;
/**
 * LogicalClient is a network client capable of delivering NetworkPayloads 
 * to a LogicalServer on another database. It supplies the database with an 
 * easy-to-use request-response network protocol.
 * 
 * 
 * @author Egil Srensen
 *
 */

public class LogicalClient implements Runnable{

	// The instance of the client
	private static LogicalClient instance;
	// The stream to write the requests to
	private ObjectOutputStream oos;
	// The stream to read responses from
	private ObjectInputStream ois;
	// The socket / connection to the LogicalServer
	private Socket networkConnection;
	// The port on the LogicalServer to connect to
	private int logServerPort = 12345;
	// Should the LogicalClient shut down?
	private boolean shutdown;
	
	/**
	 * Get the instance of the LogicalClient
	 * 
	 * @return		The running instance
	 */
	public static LogicalClient getInstance(){
		if (instance == null)
			instance = new LogicalClient();
		return instance;
	}
	
	/**
	 * Private constructor to initialize and run the client
	 *
	 */
	private LogicalClient(){
		shutdown = false;
		new Thread(this,"LogicalClient").start();
		reconnect();
	}
	
	/**
	 * Tell the client to shut down on next opportunity
	 *
	 */
	public void shutdown(){
		shutdown = true;
		instance = null;
	}
	
	
	/**
	 * Run the server
	 */
	public void run(){
		while(!shutdown){
			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}
	
	
	/**
	 * Send a NetworkPayload and return the response from the LogicalServer
	 * 
	 * @param payload		The request-payload
	 * @return				The response-payload
	 */
	public synchronized NetworkPayload send(NetworkPayload payload){
		NetworkPayload answer = null;

		// Try to send the networkpayload by using the ObjectOutputStream
		try {
			oos.reset();
			oos.writeObject(payload);
			oos.flush();			
			answer = (NetworkPayload) ois.readObject();
			
		} catch (IOException e) {
			// If there is a network problem, reconnect
			System.out.println("[warning] connection seems lost ... retrying ...");
			oos = null;
			ois = null;
			reconnect();
			return new NetworkPayload(NetworkPayload.TYPE_ERROR,"Could not be sent, Hot Standby not available");
		} catch (ClassNotFoundException e){
			// The returned class is not a NetworkPayload. This should never happen
			System.out.println("[warning] returned class not found ... " + e.getMessage());
		}
		
		return answer;
		
		
		
	}
	
	
	
	/**
	 * Reconnect to the log receiver and create the ObjectOutputStream needed 
	 * for the shipper
	 *
	 */
	private synchronized void reconnect(){
		String logServer = HotStandbyController.getInstance().getHotStandbyUrl();
		System.out.println("[status] LogShipper is connecting to: " + logServer + ":" + logServerPort);
		try {
			// Create the connection
			networkConnection = new Socket(logServer, logServerPort);
			networkConnection.setKeepAlive(true);
			oos = new ObjectOutputStream(networkConnection.getOutputStream());
			ois = new ObjectInputStream(networkConnection.getInputStream());
		} catch (ConnectException e){
			// If the connection was refused, retry after a random number of seconds between 1 to 10
			long randomTime = (int) (Math.random() * 10000);
			System.out.println("[error] Could not connect: " + e.getMessage() + ". Retrying in " + (randomTime / 1000) + " seconds");
			
			try{
				Thread.sleep(randomTime);
			}
			catch (InterruptedException e1){
				System.out.println("[alert] Thread was intterrupted ... " + e1.getMessage());
			}
			
			reconnect();
		} catch (UnknownHostException e1) {
			e1.printStackTrace();
		} catch (IOException e1) {
			e1.printStackTrace();
		}
	}

	
	
}
