package org.apache.derby.impl.services.net;

import java.io.EOFException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.ServerSocket;
import java.net.Socket;

import org.apache.derby.impl.services.hotstandby.HotStandbyController;
import org.apache.derby.impl.services.net.receiver.LogReceiver;
import org.apache.derby.impl.store.logical.log.LogicalLog;
/**
 * The Logical Server provides, along with the LogicalClient, an easy-to-use 
 * request-response protocol. The Logical Server in the net-service is responsible 
 * for answering lsn-requests, 2pc-related requests and log shipments
 * 
 * @author Egil Srensen
 *
 */
public class LogicalServer extends Thread{

	// The instance of the LogicalServer
	private static LogicalServer instance;
	// The stream to read the request from
	private ObjectInputStream ois;
	// The stream to write the response to
	private ObjectOutputStream oos;
	// The socket the client connects to
	private ServerSocket socket;
	// The port the server listend to
	private int serverPort = 12345;
	// The url of the currently connected client
	private String clientId;
	// A signal for the primary to wait for a little while
	private boolean wait;
	
	/**
	 * Initialize, create and start the LogicalServer
	 *
	 */
	public static void init(){
		if (instance == null){
			instance = new LogicalServer();
			new Thread(instance).start();
		}
	}
	
	public void setWait(boolean toggle){
		wait = toggle;
	}
	
	
	/**
	 * Get the instance of the running server
	 * 
	 * @return		The running instance
	 */
	public static LogicalServer getInstance(){
		return instance;
	}
	
	/**
	 * Empty private constructor
	 *
	 */
	private LogicalServer(){
		this.setName("LogicalServer");
		wait = false;
	}
	
	
	/**
	 * Run the server
	 */
	public void run() {
		socket = null;
		try {
			// Initialize the connection
			socket = new ServerSocket(serverPort);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		
		while (true){
			boolean running;
			Socket client = null;
			// Wait for connections
			System.out.println("[wait] Waiting for connections ...");
			try {
				client = socket.accept();
				client.setKeepAlive(true);
				clientId = client.getInetAddress().getHostAddress();
				System.out.println("[connection] Got a connection from " + clientId);
				ois = new ObjectInputStream(client.getInputStream());
				oos = new ObjectOutputStream(client.getOutputStream());
				running = true;

			} catch (IOException e) {
				e.printStackTrace();
				running = false;
			}
			// While the connection is up, receive requests, do the appropriate actions 
			// and reply to the invoker
			while (running){
				try {

					// Get the incoming request
					NetworkPayload payload = (NetworkPayload) ois.readObject();
					NetworkPayload answer;
					
					// Determine the type of the request and do the appropriate actions
					//TODO: Improve the 2-safe 2PC aspect. As of now, even though a 2PC_COMMIT_ACK has been sent there is no guarantee that the operation will be committed on the HotStandby. Even if the Primary is told that the 2PC failed, the operation might still be committed on the HotStandby. ACID is compromised in 2-safe as of now. However, this will not have priority in this project, but should be sorted out if this source code should ever be used for 2-safe replication.
					switch (payload.getType()){
					case NetworkPayload.TYPE_LOG:{
						if (wait){
							answer = new NetworkPayload(NetworkPayload.TYPE_WAIT,null);
							break;
						}
						else{
							LogicalLog newLog = (LogicalLog) payload.getPayload();
							int firstLsn = newLog.getFirst().getLsn();
							int lastLsn = newLog.getLast().getLsn();
							if (HotStandbyController.getInstance().getReplicationMode() == HotStandbyController.TWO_SAFE){
								LogReceiver.getInstance().receive(newLog);
							}
							answer = new NetworkPayload(NetworkPayload.TYPE_ACK, "ACK log: [" + firstLsn + " - " + lastLsn + "]");
//							answer = new NetworkPayload(NetworkPayload.TYPE_ACK, null);
							break;
						}
						
					}
					case NetworkPayload.TYPE_LSN_REQUEST:{
						answer = new NetworkPayload(NetworkPayload.TYPE_LSN_RESPONSE,new Integer(LogReceiver.getInstance().getNextLsn()));
						break;

					}
					case NetworkPayload.TYPE_2PC_PREP:{
//						System.out.println("Got 2pc-prep");
						answer = new NetworkPayload(NetworkPayload.TYPE_2PC_VOTE_COMMIT, null);
						break;
					}
					case NetworkPayload.TYPE_2PC_DOCOMMIT:{
//						System.out.println("Got 2pc_docommit");
						answer = new NetworkPayload(NetworkPayload.TYPE_2PC_COMMIT_ACK, null);
						break;
					}
					default:{
						answer = new NetworkPayload(NetworkPayload.TYPE_ERROR,"Unkown request: " + payload.getType());
						break;
					}
					}
					
					// Respond
					oos.reset();
					oos.writeObject(answer);
					oos.flush();
					
					if (HotStandbyController.getInstance().getReplicationMode() == HotStandbyController.ONE_SAFE && payload.getType() == NetworkPayload.TYPE_LOG && !wait){
						LogicalLog newLog = (LogicalLog) payload.getPayload();
						LogReceiver.getInstance().receive(newLog);						
					}
					
					
					
				} catch (EOFException e){
					// The connection was broken, flush to disk and go back to waiting for connections
					System.out.println("[error] EOF detected ... connection shut down ?");
					running = false;
				} catch (IOException e) {
					e.printStackTrace();
				} catch (ClassNotFoundException e){
					e.printStackTrace();
				}
			}
		}
		
		
		
	}
	
	
	
	
}
