package org.apache.derby.impl.services.net.shipper;

import java.io.IOException;

import org.apache.derby.iapi.services.context.ContextService;
import org.apache.derby.iapi.services.context.ShutdownException;
import org.apache.derby.impl.services.hotstandby.HotStandbyController;
import org.apache.derby.impl.services.net.LogicalClient;
import org.apache.derby.impl.services.net.NetworkPayload;
import org.apache.derby.impl.store.logical.log.LogicalLog;

/**
 * 
 * The log shipper is a class that is responsible for shipping 
 * the shared log between the primary database and the backup. 
 * It initiates a connection on startup, gets the next Lsn 
 * for the logical logger and keeps the connection up for the 
 * duration of the database. The log is shipped to the backup on 
 * either timeout, the log exceeding a size limit or on demand.
 * 
 * @author Egil Srensen
 *
 */

public class LogShipper implements Runnable{
	// The default queue size
	private final int DEFAULT_MAX_CUE_SIZE = 250;
	// The default timeout interval
	private final long DEFAULT_TIMEOUT_INTERVAL = 1000*10;
	// The pollingInterval, how many ms between each timeout/size check
	private long pollingInterval = 100;	
	// The selected size of the queue
	private int maxQueueSize;
	// The selected timeout interval
	private long timeOutInterval;
	// The time remaining until the next timeout occurs
	private long timeUntilNextTimeout;
	// The active LogShipper instance
	private static LogShipper instance;
	// The log shared between the shipper and the LogicalLogger
	private LogicalLog logBuffer;

	
	private LogicalLog log;
	// The next lsn
	private int nextLsn;
	// Flag to tell the shipper that the database is shutting down
	private boolean shutDown = false;
	
	/**
	 * Get the instance of the LogShipper, ther should not exist any more 
	 * than one shipper at one time.
	 * 
	 * @param logBuffer		The shared log
	 * @return				Returns an instance of the LogShipper
	 */
	public static synchronized LogShipper getInstance(LogicalLog logBuffer){
		if (instance == null){
			instance = new LogShipper(logBuffer);
		}
		return instance;
	}
	
	public static synchronized LogShipper getInstance(){
		return instance;
	}
	
	/**
	 * Creates a new LogShipper with the selected shared log
	 * 
	 * @param log		The shared logical log
	 */
	private LogShipper(LogicalLog logBuffer){
		
		// Initialize the logical log

		
//		System.out.println("Creating logshipper");
		
		
		this.logBuffer = logBuffer;
		maxQueueSize = DEFAULT_MAX_CUE_SIZE;
		timeOutInterval = DEFAULT_TIMEOUT_INTERVAL;
		timeUntilNextTimeout = timeOutInterval;
		nextLsn = 0;
		// Create and start the LogShipper as a separate thread
		
		if (! HotStandbyController.getInstance().isInHotStandbyMode()){
			log = LogicalLog.readFromDisk();
			Thread t = new Thread(this, "LogShipper");
			t.start();
		}
	}
	
	
	/**
	 * Runs the LogShipper. Checks for shutdown and timeouts/log size exceeded.
	 */
	public void run() {
		

		while (!shutDown){
			
			// Checking if the system has been/is shutting down
			try{
				ContextService.getFactory();	
			}
			catch (ShutdownException e){
				System.out.println("[status] Shutdown detected ...");
				
				// If more logs left, transmit these now
				if (logBuffer.size() != 0){
					timeUntilNextTimeout = 0;
				}
				// Tell the loop that the system has been shut down
				
				shutDown = true;

			}

			// If timeout occurred or max cue size is exceeded, ship the log
			if (timeUntilNextTimeout <= 0 || logBuffer.size() >= maxQueueSize){
				try {				
					shipLogs();
				} catch (IOException e) {
					e.printStackTrace();
				}
				// Reset the timeout-counter
				timeUntilNextTimeout = timeOutInterval;
			}
			try {
				// Wait through the polling interval (to preserve resources)
				Thread.sleep(pollingInterval);
				// Count down for timeout
				timeUntilNextTimeout -= pollingInterval;
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		// Finished, shutting down and closing all network connections
		System.out.println("[status] All logs have been sent ... shutting down shipper ...");
			System.out.println("[success] done");
		
	}
	
	
	/**
	 * Get the next lsn. If the nextLsn is not retrieved, get it from the log receiver
	 * If it has been retrieved once, increment it and use the local value instead
	 * 
	 * @return 		returns the lsn for the next loggable operation
	 */
	public synchronized int getNextLsn(){
		if (nextLsn == 0 && HotStandbyController.getInstance().isHotStandByAlive()){
//				System.out.println("[status] Getting lsn ....");
				
				NetworkPayload payload = new NetworkPayload(NetworkPayload.TYPE_LSN_REQUEST,null);
				NetworkPayload response = LogicalClient.getInstance().send(payload);
				
				if (response.getType() == NetworkPayload.TYPE_LSN_RESPONSE){
					nextLsn = ((Integer) response.getPayload()).intValue();
//					System.out.println("[success] Next lsn is: " + nextLsn);
				}
				else
					System.out.println("[error] Could not get next lsn from server");
				
				int localNextLsn;
				if (log.size() > 0)
					localNextLsn = log.getLast().getLsn() + 1;
				else
					localNextLsn = 1;
				
				if (nextLsn != localNextLsn){
					System.err.println("[local lsn] " + localNextLsn + " - [remote lsn] " + nextLsn);
					System.err.println("Selecting the largest lsn for now");
					if (nextLsn > localNextLsn){
						System.err.println("Critical error! LSN on Hot Standby is larger than local LSN. Restart the server in Hot StandbyMode to Catch Up");
						System.err.println("Quitting ...");
						System.exit(-1);
					}
					if (nextLsn < localNextLsn){
						System.err.println("Hot Standby is out of date. Engaging catchup ...");
//						provideCatchUp();
						nextLsn = localNextLsn;
					}
				}

		}
		else if (! HotStandbyController.getInstance().isHotStandByAlive() && nextLsn == 0){
			nextLsn = LogicalLog.readFromDisk().getLast().getLsn() + 1;
		}
		return nextLsn++;
	}
	
	
	/**
	 * Try to ship the log right now. Returns true if the logs were 
	 * succesfully shipped. However, if an error occurs a false is returned, but 
	 * that does not necessarily mean that the logs will never be sent, they could 
	 * just not be sent right now
	 * 
	 * @return		The result of the log shipment
	 */
	public boolean shipNow(){
		try {
			// Ship the logs which will in turn be redone on the HotStandby
			shipLogs();
			return true;
		} catch (IOException e) {
			timeUntilNextTimeout = 0;
			e.printStackTrace();
			return false;
		}
	}
	
	public boolean prepareTwoSafe(int transId){
		//TODO: To fully support 2PC make sure the commit-operation is not written to the log before it is sent at a later stage.
		// Very simple support for 2PC
		if (HotStandbyController.getInstance().getReplicationMode() == HotStandbyController.TWO_SAFE){
			// Send a prepare message
			NetworkPayload payload = new NetworkPayload(NetworkPayload.TYPE_2PC_PREP, null);
			NetworkPayload answer = LogicalClient.getInstance().send(payload);
			// If the answer is not a commit-vote, abort transaction
			if (answer.getType() != NetworkPayload.TYPE_2PC_VOTE_COMMIT){
				return false;
			}
			else{
				// If the answer is a commit-vote, send a do-commit message
				payload = new NetworkPayload(NetworkPayload.TYPE_2PC_DOCOMMIT, null);
				answer = LogicalClient.getInstance().send(payload);
				// If the answer is not ack, abort transaction
				if (answer.getType() != NetworkPayload.TYPE_2PC_COMMIT_ACK){
					return false;
				}
			}
		}

		
		
		return true;
	}
	
	
	/**
	 * Try to ship the logs now.
	 * 
	 * If the HotStandby database is not considered to be alive, the log is flushed to 
	 * disk instead of being sent to the HotStandby. When catchup is engaged, the 
	 * flushed log will be sent to the HotStandby database and when it is finally 
	 * caught up it will notify HotStandbyController that the HotStandby is alive again 
	 * and shipping will resume as normal.
	 * 
	 * @throws IOException		If the connection is broken or another ioexception is detected during the shipping the exception will be thrown to the invoker of this method.
	 */
	private synchronized void shipLogs() throws IOException{
//		System.out.println("Trying to ship ... isHotStandbyAlive ? " + HotStandbyController.getInstance().isHotStandByAlive());
		if (logBuffer.size() > 0 && HotStandbyController.getInstance().isHotStandByAlive()){
			LogicalLog flushedLog = logBuffer.flush();
			
//			System.out.println("[transmit] Sending log: [" + flushedLog.getFirst().getLsn() + " - " + flushedLog.getLast().getLsn() + "]");
			NetworkPayload payload = new NetworkPayload(NetworkPayload.TYPE_LOG, flushedLog);
			
			NetworkPayload answer = LogicalClient.getInstance().send(payload);
			
			while (answer.getType() == NetworkPayload.TYPE_WAIT){
				long retry = (long) (int) (Math.random() * 3000);
//				System.out.println("Hot Standby Signalled wait, retrying in " + (int) (retry/1000) + " seconds ...");
				try {
					Thread.sleep(retry);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				answer = LogicalClient.getInstance().send(payload);
			}
			
			if (answer.getType() == NetworkPayload.TYPE_ACK){
//				System.out.println("[result] " + answer.getPayload().toString());
				
				log.addElements(flushedLog);
				log.flushLogToDisk();
			}
		}
		// If the HotStandbyController says the HotStandby is not alive, the log is flushed to disk
		else if (logBuffer.size() > 0 && ! HotStandbyController.getInstance().isHotStandByAlive()){
//			System.out.println("[transmit] Hot Standby is not alive");
			
			LogicalLog flushedLog = logBuffer.flush();
			log.addElements(flushedLog);
			log.flushLogToDisk();
		}

	}
	

	/**
	 * Provide catchup for newly discovered HotStandby
	 *
	 */
	public synchronized void provideCatchUp(){
		int currentLocalLsn = 0;
		if (log.size() > 0)
			currentLocalLsn = log.getLast().getLsn();
		
		// Start the catchupservice
		LogicalCatchUp catchUpService = new LogicalCatchUp(++currentLocalLsn);
	}
	
	protected void setBuffer(LogicalLog buffer){
		logBuffer = buffer;
	}

	/**
	 * Set the max queue size
	 * 
	 * @param size		The new max size for the queue
	 */
	public synchronized void setMaxQueueSize(int size){
		maxQueueSize = size;
	}
	
	/**
	 * Set the timeout interval
	 * 
	 * @param interval	The new timeout interval value
	 */
	public synchronized void setTimeOutInterval(long interval){
		timeOutInterval = interval;
		timeUntilNextTimeout = timeOutInterval;
	}
	
}
