package org.apache.derby.impl.services.net.shipper;

import java.util.ArrayList;

import org.apache.derby.impl.services.hotstandby.HotStandbyController;
import org.apache.derby.impl.services.net.LogicalClient;
import org.apache.derby.impl.services.net.NetworkPayload;
import org.apache.derby.impl.store.logical.log.LogicalLog;
/**
 * LogicalCatchUp is a service created to provide catchup to newly discovered 
 * HotStandby databases. The LogicalCatchup service checks if a catchup is needed
 * between the two databases. If it is needed the log is scanned for changes and sent 
 * to the HotStandby until it is finally caught up.
 * 
 * Some severe problems occur during catchup if the database is under a lot of 
 * stress. The log can fill faster than the catchup can provide logs for the 
 * HotStandby and the primary database will sooner or later run out of memory. This 
 * could be solved by denying or limiting connections while the catchup is in 
 * progress. However, this is beyond the scope of this project.
 * 
 * @author Egil Srensen
 *
 */
public class LogicalCatchUp extends Thread{
	// The last lsn from the local database
	private int currentLocalLsn;
	
	/**
	 * Create and start a new LogicalCatchUp service
	 * 
	 * @param currentLocalLsn		The last lsn of the local database
	 */
	public LogicalCatchUp(int currentLocalLsn){
		super();
		this.currentLocalLsn = currentLocalLsn;
		this.setName("LogicalCatchUp");
//		System.out.println("Creating logical catch up - service");
		start();
	}
	
	/**
	 * Run the service, consists of calling the provideCatchUp()-method
	 */
	public void run(){
		provideCatchUp();
	}
	
	
	/**
	 * Provide catchup for the HotStandby. The log is scanned for changes after the 
	 * last remote lsn reported and all log records with a higher lsn than that are 
	 * shipped to the HotStandby.
	 * 
	 * The local log is read from disk, compared to the remote lsn reported and split 
	 * into chunks of 3000 log records who are in turn sent to the HotStandby. This step 
	 * is repeated until all of the log residing on the disk is shipped. (This is needed 
	 * as the log on disk can be changed while the catchup is in progress).
	 * 
	 * Finally the service notifies the HotStandbyController that the HotStandby is now 
	 * alive again and normal LogShipping is resumed.
	 *
	 */
	private void provideCatchUp(){
		System.out.println("[status] Providing catch up for the arrived Hot Standby");
		LogicalClient client = LogicalClient.getInstance();
		NetworkPayload payload = new NetworkPayload(NetworkPayload.TYPE_LSN_REQUEST,null);
		
		NetworkPayload answer = client.send(payload);
		while (answer.getType() == NetworkPayload.TYPE_ERROR){
			answer = client.send(payload);
		}
		// Check if catchup is needed
		if (answer.getType() == NetworkPayload.TYPE_LSN_RESPONSE){
			int lastRecordedLsn = ((Integer) answer.getPayload()).intValue();
			if (lastRecordedLsn < currentLocalLsn){
				System.out.println("Need to redo everything from " + lastRecordedLsn);
				LogicalLog diskLog = LogicalLog.readFromDisk();
				System.out.println("Last log on disk: " + diskLog.getLast().getLsn());
				LogicalLog catchupLog = diskLog.split(lastRecordedLsn);
				System.out.println("Last split lsn: " + catchupLog.getLast().getLsn());
				// As long as the log on the disk is not caught up, 
				// do the catchup-routine again
//				while (catchupLog.size() > 0){
					ArrayList logFragments = catchupLog.divide();
					System.out.println("Log fragments : " + logFragments.size());
					// Ship the log fragments to the HotStandby
					while (logFragments.size() > 0){
						LogicalLog sendLog = (LogicalLog) logFragments.remove(0);
						System.out.println("[transmit] Sending log: [" + sendLog.getFirst().getLsn() + " - " + sendLog.getLast().getLsn() + "]");
						payload = new NetworkPayload(NetworkPayload.TYPE_LOG, sendLog);

						answer = LogicalClient.getInstance().send(payload);
						
						while (answer.getType() == NetworkPayload.TYPE_WAIT){
							long retry = (long) (int) (Math.random() * 3000);
//							System.out.println("Hot Standby Signalled wait, retrying in " + retry);
							try {
								Thread.sleep(retry);
							} catch (InterruptedException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}
							answer = LogicalClient.getInstance().send(payload);

						}
						
						System.out.println("[result] " + answer.getPayload().toString());
						lastRecordedLsn = sendLog.getLast().getLsn();
					}
					//TODO: Broken her, m fikses grunnet ny logreceiver-rutine/synch
//					diskLog = LogicalLog.readFromDisk();
//					catchupLog = diskLog.split(lastRecordedLsn + 1);
//				}
			}
			else if(lastRecordedLsn > currentLocalLsn){
				System.err.println("Critical error! LSN on Hot Standby is larger than local LSN. Restart the server in Hot StandbyMode to Catch Up");
				System.exit(-1);
			}
			else {
				System.out.println("System is already up-to-date");
			}
		}
		// Catchup is finished
		System.out.println("[done]");
		// Notify the HotStandbyController
		HotStandbyController.getInstance().setHotStandByAlive(true);
	}
	
}
