package org.apache.derby.impl.store.logical.log;

import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OptionalDataException;
import java.io.Serializable;
import java.util.ArrayList;

/**
 * 
 * A class implementing a classical linked list of log elements. 
 * It is serializable so that it can be sent through a outputstream and 
 * cloneable because of the need to avoid race conditions.
 * 
 * The Logical Log Records are encased in LogElements to add the support 
 * of nextElement needed in the linked list.
 * 
 * The logical log has a fixed maximum size and functions as a circular buffer
 * 
 * @author Egil Srensen
 *
 */
public class LogicalLog implements
		org.apache.derby.iapi.store.raw.log.LogicalLog, Cloneable {


	private static final long serialVersionUID = 1L;


	// The maximum size of the log
	private final int MAX_SIZE = 20000;
	
	// A reference to the current element, needed to traverse the linked list
	// using hasMoreElements() and getNext()
	private transient LogElement currentElement;
	
	// The first and last elements in the linked list
	private transient LogElement firstElement;
	private transient LogElement lastElement;
	
	// The relative path to where the logicallogs are to be written
	private static String logFolder = "logicallog/";
	// The preset for the logical log files
	private static String logFile = logFolder + "logical.log";
	// The name of the file with the nextLsn written
	private static String lsnFile = logFolder + "logicalLsn.log";
	
	// The size of the log
	private int size;
	
	/**
	 * Standard constructor, creates an empty LogicalLog with no elements.
	 *
	 */
	public LogicalLog(){
		firstElement = null;
		currentElement = null;
		lastElement = null;
		size = 0;
	}
	
	
	/**
	 * Constructor with a logical log record as a parameter. Creates 
	 * a LogicalLog with the provided log record as the first and 
	 * last object.
	 * 
	 * Using this constructor is the same as doing:
	 * <code>
	 * LogicalLog log = new LogicalLog();
	 * log.add(lr);
	 * </code>
	 *  
	 * @param lr 		The Logical Log Record to be added to the new Logical Log
	 */
	public LogicalLog(LogicalLogRecord lr){
		LogElement element = new LogElement(lr);
		firstElement = element;
		currentElement = element;
		lastElement = element;
		size = 1;
	}
	
	
	/**
	 * flush() is a method added to remove race conditions between 
	 * the add() method and the need to empty the Logical Log after 
	 * it has been sent using the LogShipper. The current Logical Log 
	 * is cloned and returned and this Logical Log is then emptied and 
	 * ready for further use.
	 * 
	 * @see 		org.apache.derby.impl.services.net.shipper.LogShipper
	 * @return		Returns a clone of the current Logical Log
	 * 
	 */
	public LogicalLog flush(){
		LogicalLog purgedLog = null;
		try {
			purgedLog = (LogicalLog) this.clone();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
		}
		currentElement = null;
		firstElement = null;
		lastElement = null;
		size = 0;
		return purgedLog;
	}
	
	
	/**
	 * Overridden method to write the object to the stream in an iterative manner 
	 * instead of javas default recursive marshalling
	 * 
	 * @param out				The stream to write to
	 * @throws IOException
	 */
	private void writeObject(ObjectOutputStream out) throws IOException{
		out.defaultWriteObject();
		currentElement = firstElement;
		while (hasMoreElements()){
			out.writeObject(getNext());
		}
		out.flush();
	}
	
	
	/**
	 * Overridden method to read the object from a stream in an iterative manner
	 * instead of javas default recursive marshalling
	 * 
	 * @param in				The stream to read the log from
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException{
		in.defaultReadObject();
		firstElement = null;
		currentElement = null;
		lastElement = null;
		size = 0;
		boolean finished = false;
		LogicalLogRecord logElement = (LogicalLogRecord) in.readObject();
		while (!finished){
			try{
				add(logElement);
				logElement = (LogicalLogRecord) in.readObject();
			}
			catch (OptionalDataException e){
				finished = true;
			}
			
		}
	}
	


	/**
	 * Adds a Logical Log Record to the Logical Log. If the lsn of the 
	 * log to be added is <= than the last lsn recorded in the log it is
	 * not added. This is to ensure that the log is written in a strict 
	 * incremental way
	 * 
	 * @param lr		The Logical Log Record to be added.
	 */
	public synchronized void add(LogicalLogRecord lr) {
		LogElement element = new LogElement(lr);
		
		if (size == 0){
			currentElement = element;
			firstElement = element;
			lastElement = element;
			size++;
		}
		else{
			if (lr.getLsn() > lastElement.getLogRecord().getLsn()){
				element.setPrevElement(lastElement);
				lastElement.setNextElement(element);
				lastElement = element;
				size++;
			}
			else{
				System.out.println("Log Record #" + lr.getLsn() + " was not added, exists in log already");
			}
			// Provide a circular buffer with size MAX_SIZE
			if (size > MAX_SIZE){
				firstElement = firstElement.nextElement;
				firstElement.setPrevElement(null);
				size--;
				reset();
			}
		}
		

	}

	
	/**
	 * Adds an entire logical log to this log, used in LogReceiver when 
	 * an entire log is received and then added to the current log. The log 
	 * is traversed and each log record is added one by one.
	 * 
	 * @see org.apache.derby.impl.services.net.receiver.LogReceiver
	 * @param log		The logical log to be added to this logical log
	 */
	public void addElements(org.apache.derby.iapi.store.raw.log.LogicalLog log) {
		log.reset();
		while (log.hasMoreElements()){
			LogicalLogRecord lr = log.getNext();
			add(lr);
		}

	}

	
	/**
	 * Get the first log element of the Logical Log
	 * 
	 * @return		The first LogicalLogRecord of this log
	 */
	public LogicalLogRecord getFirst() {
		return firstElement.getLogRecord();
	}

	/**
	 * Get the last log element of the Logical Log
	 * 
	 * @return		The last LogicalLogRecord of this log
	 */
	public LogicalLogRecord getLast() {
		return lastElement.getLogRecord();
	}

	/**
	 * Get the next (if any) log element of the Logical Log
	 * 
	 * @return		The next LogicalLogRecord of this log
	 */
	public LogicalLogRecord getNext() {
		LogicalLogRecord nextLogRecord = currentElement.getLogRecord();
		
		currentElement = currentElement.getNextElement();
		return nextLogRecord;
	}
	
	
	/**
	 * Rewind the log to a selected lsn and returns this log record.
	 * Added to enhance the scan needed during a transaction-abort.
	 * 
	 * @param lsn		The lsn of the log to be found
	 * @return			The log record corresponding to the lsn
	 */
	public LogicalLogRecord rewind(int lsn){
		
		if (firstElement.getLogRecord().getLsn() < lsn){
			LogElement watermark = lastElement;
			while (watermark.getPrevElement() != null){
				if (watermark.getLogRecord().getLsn() == lsn)
					return watermark.getLogRecord();
				else
					watermark = watermark.getPrevElement();
			}
		}
		return null;
	}

	
	/**
	 * Checks if this Logical Log has more elements
	 * 
	 * @return		boolean
	 */
	public boolean hasMoreElements() {
		return (currentElement != null);
	}

	
	/**
	 * Gets the size of this Logical Log
	 * 
	 * @return		The size of the log
	 */
	public int size() {
		return size;
	}
	
	
	/**
	 * Writes the log to screen, written for debug purposes
	 */
	
	public String toString(){
		
		String returned = "Logical Log: \n\n";
		
		LogElement oldCurrentElement = currentElement;
		
		currentElement = firstElement;
		
		while (hasMoreElements()){
			returned += getNext();
			returned += "\n";
		}
		
		currentElement = oldCurrentElement;
		
		return returned;
	}
	
	
	
	
	/**
	 * Flushes the logical log to disk. Every flush writes to a new log file with the 
	 * lsn of the first log record as the files extension
	 *
	 */
	public synchronized void flushLogToDisk(){
		if (size() > 0){
//			System.out.println("[write] Flushing log to disk ...");
			try {
				ObjectOutputStream fos = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(logFile, false)));
				fos.writeObject(this);
				fos.flush();
				fos.close();
				fos = null;
			}
			catch (IOException e1){
				e1.printStackTrace();
			}
			
			
//			System.out.println("[write] done!");
			
		}
	}
	

	/**
	 * Reset pointers in the log, moving the currentElement-pointer to firstElement
	 */
	public void reset(){
		currentElement = firstElement;
	}
	
	
	/**
	 * Split the log around a chosen lsn. Returns a log with all log records 
	 * newer than (with a higher lsn than) the lsn provided.
	 * 
	 * @param lsn		The lsn to split the log on
	 * @return			The last part of the splitted list
	 */
	public LogicalLog split(int lsn){
		if (lsn >= getFirst().getLsn() && lsn <= getLast().getLsn()){
			LogElement pointer = firstElement;
			LogicalLog returned = new LogicalLog();
			while (pointer != null){
				if (pointer.getLogRecord().getLsn() >= lsn)
					returned.add(pointer.getLogRecord());
				pointer = pointer.getNextElement();
			}
			return returned;
		}
		return new LogicalLog();
	}
	
	
	
	/**
	 * Divide the log into an ArrayList with logs with a maximum size of 3000.
	 * This method is needed by the catchupservice to ensure that the logs 
	 * sent to the new HotStandby are not too large.
	 * 
	 * @return		An ArrayList of logs with a max size of 3000
	 */
	public ArrayList divide(){
		
		LogElement pointer = firstElement;
		LogicalLog returned = new LogicalLog();
		ArrayList returnedList = new ArrayList();
		
		while (pointer != null){
			returned.add(pointer.getLogRecord());
			
			if (returned.size() == 3000 || pointer.getNextElement() == null){
				try {
					returnedList.add(returned.clone());
				} catch (CloneNotSupportedException e) {
					e.printStackTrace();
				}
				returned = new LogicalLog();
			}
			
			pointer = pointer.getNextElement();
		}
		
		return returnedList;
	}

	
	/**
	 * Read the log from disk. The log to be read needs to be located in the folder
	 * logicallog/logical.log
	 * 
	 * @return		The logical log read from disk
	 */
	
	public static synchronized LogicalLog readFromDisk(){
		LogicalLog log = null;
		try {
			log = (LogicalLog) new ObjectInputStream(new FileInputStream(logFile)).readObject();
		} catch (FileNotFoundException e){
			log = new LogicalLog();
		} catch (ClassNotFoundException e){
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
//		System.out.println("Read log file: " + log.size + " from disk");
		return log;
	}
	
	
	
	/**
	 * 
	 * The LogElement class is a wrapper around the LogicalLogRecord 
	 * to be shipped to another process.
	 * 
	 * @author Egil Srensen
	 *
	 */
	private class LogElement implements Serializable{

		// Automatically generated serial version id
		private static final long serialVersionUID = -987066189794844222L;
		// The contents of this logElement
		private LogicalLogRecord lr;
		// A pointer to the next element
		private LogElement nextElement;
		// A pointer to the previous element
		private LogElement prevElement;
		
		/**
		 * Standard constructor, creates a log element containing 
		 * a LogicalLogRecord
		 * 
		 * @param lr		The LogicalLogRecord to be added to the LogElement
		 */
		public LogElement(LogicalLogRecord lr){
			this.lr = lr;
			nextElement = null;
			prevElement = null;
		}
		

		/**
		 * Set the next element
		 * 
		 * @param nextElement	The next element of this LogElement
		 */
		public void setNextElement(LogElement nextElement){
			this.nextElement = nextElement;
		}
		
		/**
		 * Get the next element. Used to traverse the linked list
		 * 
		 * @return		The next (if any) element
		 */
		public LogElement getNextElement(){
			return nextElement;
		}
		
		
		/**
		 * Get the log record inside this log element
		 * 
		 * @return		The LogicalLogRecord for this element
		 */
		
		public LogicalLogRecord getLogRecord(){
			return lr;
		}

		/**
		 * Get the previous element. Used to traverse the linked list
		 * 
		 * @return		The previous (if any) element
		 */
		public LogElement getPrevElement() {
			return prevElement;
		}

		/**
		 * Set the previous element
		 * 
		 * @param prevElement	The previous element of this LogElement
		 */
		public void setPrevElement(LogElement prevElement) {
			this.prevElement = prevElement;
		}
	}

}
