package org.apache.derby.impl.store.logical.log;
import java.util.Hashtable;

import org.apache.derby.iapi.error.StandardException;
import org.apache.derby.iapi.sql.dictionary.ColumnDescriptor;
import org.apache.derby.iapi.types.DataValueDescriptor;
import org.apache.derby.impl.services.hotstandby.HotStandbyController;
import org.apache.derby.impl.services.net.shipper.LogShipper;
import org.apache.derby.impl.store.logical.operations.LogicalImage;
import org.apache.derby.impl.store.logical.operations.LogicalOperation;
import org.apache.derby.impl.store.logical.operations.LogicalOperationFactory;

/**
 * The logical logger is created as a singleton with three 
 * differene log services: log, logAndCommit and logAndAbort
 * 
 * 
 * @author Egil Srensen
 *
 */

public class LogicalLogger {
	
	// The instance of the Logical Logger
	private static LogicalLogger instance;
	// A transaction table used to get the last lsn from 
	// the active transaction
	private Hashtable transactionTable;
	// The LogicalLog
	private LogicalLog log;
	// The LogShipper
	private LogShipper ls;
	// The mode set, it can be either ONE_SAFE or TWO_SAFE
//	private int mode;
	private int nextLsn;
	
	
	/**
	 * Gets an instance of the LogicalLogger. If one exists, return it. 
	 * If not create an instance. Ensures that there are no more than 
	 * one LogicalLogger at any given time.
	 * 
	 * @return		The LogicalLogger instance
	 */
	public synchronized static LogicalLogger getInstance(){
		if (instance == null)
			instance = new LogicalLogger();
		return instance;
	}
	
	/**
	 * Private constructor used by the getInstace-method, 
	 * initiates a new logical logger
	 *
	 */
	
	private LogicalLogger(){
		System.out.println("Creating logical logger");
		transactionTable = new Hashtable();
		log = new LogicalLog();
		ls = LogShipper.getInstance(log);
		nextLsn = -1;
	}
	
	
	
	/**
	 * Main log method, creates a logical log record and adds it to 
	 * the logical log
	 * 
	 * 
	 * @param oldValue			the oldValue of the affected row
	 * @param newValue			the newValue of the affected row
	 * @param oldKey			the old Key
	 * @param newKey			the new Key
	 * @param tableName			the name of the affected table
	 * @param columns			a list of columns for this table
	 * @param transactionId		the id of the active transaction
	 * @param operationType		the operation type
	 */
	public synchronized void log(
			DataValueDescriptor[] oldValue, 
			DataValueDescriptor[] newValue,
			DataValueDescriptor oldKey,
			DataValueDescriptor newKey,
			String tableName,
			ColumnDescriptor[] columns,
			int keyColumnName,
			int transactionId, 
			int operationType
			){
		
		
		// Get the next lsn from the logShipper. The method is synchronized to 
		// ensure that the log is not filled too quickly. Some optimizations here
		// would help throughput and response times a great deal.
		nextLsn = ls.getNextLsn();
		int prevTransLsn = -1;
		
		// Check if the transaction is in the transaction table, if it is 
		// get the previousTransLsn for this transaction
		if (transactionTable.containsKey(new Integer(transactionId)))
			prevTransLsn = ((Integer) transactionTable.get(new Integer(transactionId))).intValue();
		
		// If not, add the transaction to the transactionTable
		transactionTable.put(new Integer(transactionId), new Integer(nextLsn));
		
		// Create images for the old and new image
		LogicalImage oldImage = new LogicalImage(columns, oldValue, keyColumnName, oldKey, operationType);
		LogicalImage newImage = new LogicalImage(columns, newValue, keyColumnName, newKey, operationType);

		// Create a logical operation and add the logical log record to the log
		try {
			LogicalOperation op = LogicalOperationFactory.getInstance().makeOperation(oldImage, newImage, oldKey, newKey, tableName, operationType);
			LogicalLogRecord record = new LogicalLogRecord(nextLsn,prevTransLsn,transactionId,op);
			log.add(record);
		} catch (StandardException e) {
			System.out.println("We got an exception: " + e.getErrorCode() + ": " + e.getMessage());
			e.printStackTrace();
		}		
		
		
	}
	
	/**
	 * Writes a commit log record. If the mode is set to TWO_SAFE 
	 * the log is attempted flushed to the log receiver. Returns a 
	 * boolean result depending on the success of the log flush. 
	 * If the mode is set to ONE_SAFE this always returns true, if it 
	 * is set to TWO_SAFE however if it returns true the log was shipped 
	 * succesfully, if an error occurs however the log might have been 
	 * sent, but we cannot guarantee this and the response is false.
	 * 
	 * @param transactionId		The id of the transaction to be committed
	 * @return					The result of the log shipment
	 */
	public synchronized boolean logAndCommit(
			int transactionId 
	){
		
		boolean result;
		transactionTable.remove(new Integer(transactionId));
		
		if (HotStandbyController.getInstance().getReplicationMode() == HotStandbyController.TWO_SAFE){

			result = ls.prepareTwoSafe(transactionId);
			
			if (result){
				log (null, null, null, null, null, null, 0, transactionId, LogicalOperation.COMMIT_OPERATION);
				// Send until succesful
				while (! ls.shipNow());
			}
			else{
				log (null, null, null, null, null, null, 0, transactionId, LogicalOperation.ABORT_OPERATION);
			}
		}
		else{
			log (null, null, null, null, null, null, 0, transactionId, LogicalOperation.COMMIT_OPERATION);
			result = true;
		}
		return result;

	}
	
	
	
	/**
	 * Writes an abort log record. If the mode is set to TWO_SAFE 
	 * the log is attempted flushed to the log receiver. Returns a 
	 * boolean result depending on the success of the log flush. 
	 * If the mode is set to ONE_SAFE this always returns true, if it 
	 * is set to TWO_SAFE however if it returns true the log was shipped 
	 * succesfully, if an error occurs however the log might have been 
	 * sent, but we cannot guarantee this and the response is false.
	 * 
	 * @param transactionId		The id of the transaction to be committed
	 * @return					The result of the log shipment
	 */	
	public synchronized boolean logAndAbort(
		int transactionId 
		){
		log (null, null, null, null, null, null, 0, transactionId, LogicalOperation.ABORT_OPERATION);
		
//		if (HotStandbyController.getInstance().getReplicationMode() == HotStandbyController.TWO_SAFE){
//			System.out.println("[abort] TWO-SAFE");
//			boolean result = ls.shipNow();
//			System.out.println("[TWO-SAFE result] " + result);
//			return result;
//		}
//		else{
//			System.out.println("[abort] ONE-SAFE");
			return true;
//		}		
		
	}
	
//	public LogicalLog getLog(){
//		return log;
//	}

//	public int getMode() {
//		return mode;
//	}
	


}
