package org.apache.derby.impl.store.logical.operations;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;

import org.apache.derby.iapi.store.logical.log.LogicalImage;
import org.apache.derby.iapi.store.logical.log.LogicalLoggable;
import org.apache.derby.iapi.types.DataValueDescriptor;

/**
 * A class representing the logical operation done in 
 * a logical log record
 * 
 * @author Egil Srensen
 *
 */
public abstract class LogicalOperation implements
		LogicalLoggable {
	
	// The before and after images of the affected row
	protected LogicalImage beforeImage;
	protected LogicalImage afterImage;
	
	// The old and new key for the affected row
	protected DataValueDescriptor oldKey;
	protected DataValueDescriptor newKey;
	
	// The name of the affected table
	protected String tableName;
	
	
	protected String redoStmt;
	protected String undoStmt;
	
	
	
	/**
	 * Constructor to create a logical operation
	 * 
	 * @param beforeImage		the before image of the affected row
	 * @param afterImage		the after image of the affected row
	 * @param oldKey			the old key of the affected row
	 * @param newKey			the new key of the affected row
	 * @param tableName			the name of the affected table
	 */
	public LogicalOperation(LogicalImage beforeImage, LogicalImage afterImage, DataValueDescriptor oldKey, DataValueDescriptor newKey, String tableName){
		this.beforeImage = beforeImage;
		this.afterImage = afterImage;
		this.oldKey = oldKey;
		this.newKey = newKey;
		this.tableName = tableName;
	}

	public DataValueDescriptor getNewKey() {
		return newKey;
	}

	public DataValueDescriptor getOldKey() {
		return oldKey;
	}

	public LogicalImage getRedoImage() {
		return afterImage;
	}

	public String getTableName() {
		return tableName;
	}

	public LogicalImage getUndoImage() {
		return beforeImage;
	}

	public int group() {
		return 0;
	}

	public boolean isRedoable() {
		return true;
	}

	public boolean isUndoable() {
		return true;
	}
	
	

	
	
	/**
	 * Redo this operation
	 */
	public int redoMe(Connection conn) throws SQLException {

		if (isRedoable()){
			Statement s = conn.createStatement();
			int affectedRows = s.executeUpdate(redoStmt);
			s.close();
			return affectedRows;
		}
		else
			return -1;
	}

	/**
	 * Undo this operation
	 */
	public int undoMe(Connection conn) throws SQLException {
		if (isUndoable()){
			Statement s = conn.createStatement();
			int affectedRows = s.executeUpdate(undoStmt);
			s.close();
			return affectedRows;
		}
		else
			return -1;
	}	
	
	
	public String toString(){
		String type = "";
		
		switch (group()){
			case INSERT_OPERATION:
				type = "Insert ";
				break;
			case UPDATE_OPERATION:
				type = "Update ";
				break;
			case DELETE_OPERATION:
				type = "Delete ";
				break;
			case COMMIT_OPERATION:
				type = "Commit ";
				break;
			case ABORT_OPERATION:
				type = "Abort ";
				break;
			default:
				type = "Unknown ";
				break;
		}
		type += "operation";
				
		String returned = type + "\n" + 
		"Table: " + tableName + "\n" + 
		"Old Key: " + oldKey + "\n" + 
		"New Key: " + newKey + "\n";
		
		returned += "Before image: \n" + beforeImage + "\n" + 
		"After image: \n" + afterImage + "\n";
		
		return returned;
	}




}
