package org.apache.derby.impl.store.logical.operations;

import java.sql.SQLException;

import org.apache.derby.iapi.error.PublicAPI;
import org.apache.derby.iapi.error.StandardException;
import org.apache.derby.iapi.store.logical.log.LogicalImage;
import org.apache.derby.iapi.types.DataValueDescriptor;

/**
 * The LoicalOperationFactory is used to create LogicalOperations according 
 * to their type (UPDATE, INSERT, DELETE, COMMIT or ABORT)
 * 
 * @author Egil Srensen
 *
 */
public class LogicalOperationFactory {
	// The LogicalOperationFactory instance
	private static LogicalOperationFactory instance;
	
	/**
	 * Gets the active LogicalOperationFactory. Ensures that there is only one 
	 * active LogicalOperationFactory at all times.
	 * 
	 * @return		The active logicalOperationFactory
	 */
	public static LogicalOperationFactory getInstance(){
		if (instance == null)
			instance = new LogicalOperationFactory();
		return instance;
	}
	
	/**
	 * Creates an empty LogicalOperationFactory, used by getInstance()
	 *
	 */
	private LogicalOperationFactory(){
		
	}
	
	/**
	 * Makes a LogicalOperation using the providing information
	 * 
	 * @param beforeImage			the before image of the operation
	 * @param afterImage			the after image of the operation 
	 * @param oldKey				the old key of the affected row
	 * @param newKey				the new key of the affected row
	 * @param tableName				the name of the affected table
	 * @param type					the type of the operation
	 * @return						returns a new LogicalOperation of the correct type
	 * @throws StandardException	if the provided type is not recognized a StandardException is thrown
	 */
	public LogicalOperation makeOperation(LogicalImage beforeImage, LogicalImage afterImage, DataValueDescriptor oldKey, DataValueDescriptor newKey, String tableName, int type) throws StandardException{
		
		// The type is checked and an operation of the correct type is created and returned
		switch (type){
		case LogicalOperation.UPDATE_OPERATION:
			return new LogicalUpdateOperation(beforeImage, afterImage, oldKey, newKey, tableName);
		case LogicalOperation.DELETE_OPERATION:
			return new LogicalDeleteOperation(beforeImage, afterImage, oldKey, newKey, tableName);
		case LogicalOperation.INSERT_OPERATION:
			return new LogicalInsertOperation(beforeImage, afterImage, oldKey, newKey, tableName);
		case LogicalOperation.COMMIT_OPERATION:
			return new LogicalCommitOperation(beforeImage, afterImage, oldKey, newKey, tableName);
		case LogicalOperation.ABORT_OPERATION:
			return new LogicalAbortOperation(beforeImage, afterImage, oldKey, newKey, tableName);
		}
		try {
			throw PublicAPI.wrapStandardException(StandardException.newException("Invalid logical operation type defined: " + type));
		} catch (SQLException e) {
			e.printStackTrace();
		}
		return null;

	}
	
	

	
	
}
